/* ### cpcn-1 */
/***********************************************/
/*      small-c:PC compiler                    */
/*        by Ron Cain                          */
/*  modified by CAPROCK SYSTEMS for IBM PC     */
/*  switch from Mike Bernson 1981              */
/*  adapted to core/74 cpu Ben Franchuk 2020   */
/*  input files must have only \n as line ends */
/***********************************************/

#define VERSION "Small C 1.0"
/*  Define system dependent parameters  */
/*  Stand-alone definitions         */
/*  UNIX definitions (if not stand-alone)   */
/* 16 bit cpu */
#define BITWRD 8
#define BMASK 255
#define KDIV  10000
/* KDIV 10000 BOOTSTRAP 100000 SELF COMPILE */
/* edit kdiv in assember file */
/*  20 bit cpu
#define BITWRD 10
#define BMASK 1023
#define KDIV  10000
#include bios.h
*/
#include  stdio.h
/*  Define the symbol table parameters  */
/*
#define symsiz  14
#define symtbsz 5040
#define numglbs 300
#define startglb symtab
#define endglb  startglb+numglbs*symsiz
#define startloc endglb+symsiz
#define endloc  symtab+symtbsz-symsiz
*/
/* ~ 400 symbols */
#define symsiz  14
#define symtbsz 5600
#define numglbs 350
#define startglb symtab
#define endglb   symtab+4900
#define startloc symtab+4914
#define endloc   symtab+symtbsz-symsiz
/*  Define symbol table entry format    */
#define name    0
#define ident   9
#define type    10
#define storage 11
#define offset  12
/*  System wide name size (for symbols) */
#define namesiz  9
#define namemax  8
/*  Define possible entries for "ident" */
#define variable 1
#define array   2
#define pointer 3
#define function 4
/*  Define possible entries for "type"  */
#define cchar   1
#define cint    2
/*  Define possible entries for "storage"   */
#define statik  1
#define stkloc  2
/*  Define the "while" statement queue  */
#define wqtabsz 100
#define wqsiz   4
#define wqmax   wq+wqtabsz-wqsiz
/*  Define entry offsets in while queue */
#define wqsym   0
#define wqsp    1
#define wqloop  2
#define wqlab   3
/*  Define the literal pool         */
#define litabsz 2500
#define litmax  litabsz-1
/*  Define the input line           */
#define linesize 128
#define linemax linesize-1
#define mpmax   linemax
/*  Define the macro (define) pool      */
#define macqsize 2000
#define macmax  macqsize-1
/* define switch */
#define maxsw 48
/*  Define statement types (tokens)     */
#define stif    1
#define stwhile 2
#define streturn 3
#define stbreak 4
#define stcont  5
#define stasm   6
#define stexp   7
#define stswitch  8
/* Define how to carve up a name too long for the assembler */
#define asmpref 7
#define asmsuff 7
/*  Now reserve some storage words      */
char    symtab[symtbsz];    /* symbol table */
char    *glbptr,*locptr;    /* ptrs to next entries */
int wq[wqtabsz];        /* while queue */
int *wqptr;         /* ptr to next entry */
char    litq[litabsz];      /* literal pool */
int litptr;         /* ptr to next entry */
char    macq[macqsize];     /* macro string buffer */
int macptr;         /* and its index */
char    line[linesize];     /* parsing buffer */
char    mline[linesize];    /* temp macro buffer */
int lptr,mptr;      /* ptrs into each */
/*  Misc storage    */
int nxtlab,     /* next avail label # */
litlab,     /* label # assigned to literal pool */
Zsp,        /* compiler relative stk ptr */
Lsp,        /* used for pop/push    */
Acc,        /* 0 'A' 1 'X'          */
argstk,     /* function arg sp */
ncmp,       /* # open compound statements */
errcnt,     /* # errors in compilation */
errstop,    /* stop on error            gtf 7/17/80 */
eof,        /* set non-zero on final input eof */
input,      /* iob # for input file */
output,     /* iob # for output file (if any) */
input2,     /* iob # for "include" file */
ctext,      /* non-zero to intermix c-source */
cmode,      /* non-zero while parsing c-code */
/* zero when passing assembly code */
lastst,     /* last executed statement type */
saveout,    /* holds output ptr when diverted to console       */
/*                  gtf 7/16/80 */
fnstart,    /* line# of start of current fn.    gtf 7/2/80 */
lineno,     /* line# in current file        gtf 7/2/80 */
infunc,     /* "inside function" flag       gtf 7/2/80 */
savestart,  /* copy of fnstart "    "       gtf 7/16/80 */
saveline,   /* copy of lineno  "    "       gtf 7/16/80 */
saveinfn,   /* copy of infunc  "    "       gtf 7/16/80 */
page;       /* page zero */
char   *currfn,     /* ptr to symtab entry for current fn.  gtf 7/17/80 */
*savecurr;   /* copy of currfn for #include      gtf 7/17/80 */
char    quote[2];   /* literal string for '"' */
char    *cptr;      /* work ptr to any char buffer */
int *iptr;      /* work ptr to any int buffer */
/*  >>>>> start cc1 <<<<<<      */
/*                  */
/*  Compiler begins execution here  */
/*                  */
main()
{
output=0;	/* start console */
hello();    /* greet user */
see();      /* determine options */
page= 500;  /* zero page count down */
openin();   /* first file to process */
while (input!=0)    /* process user files till he quits */
{
glbptr=startglb;    /* clear global symbols */
locptr=startloc;    /* clear local symbols */
wqptr=wq;       /* clear while queue */
macptr=     /* clear the macro pool */
litptr=     /* clear literal pool */
Zsp =       /* stack ptr (relative) */
errcnt=     /* no errors */
0;
eof=        /* not end-of-file yet */
input2=     /* no include file */
saveout=    /* no diverted output */
ncmp=       /* no open compound states */
lastst=     /* no last statement yet */
0;
fnstart=    /* current "function" started at line 0 gtf 7/2/80 */
lineno=     /* no lines read from file      gtf 7/2/80 */
infunc=     /* not in function now          gtf 7/2/80 */
quote[1]=
0;      /*  ...all set to zero.... */
quote[0]='"';       /* fake a quote literal */
currfn=NULL;    /* no function yet          gtf 7/2/80 */
cmode=1;
/* enable preprocessing  */
litlab=getlabel();
openout();
header();
parse();
if (ncmp) error("missing closing bracket");
dumpublics();
trailer();
closeout();
errorsummary();
/* openin(); no need to read more if output closed bf */
}
}
/* ### cpcn-2 */
/*                  */
/*  Abort compilation       */
/*      gtf 7/17/80     */
aborting()
{
if(input2)
endinclude();
if(input)
fclose(input);
closeout();
toconsole();
pl("Compilation aborted.");  nl();
exit();
/* end abort */}
/*                  */
/*  Process all input text      */
/*                  */
/* At this level, only static declarations, */
/*  defines, includes, and function */
/*  definitions are legal...    */
parse()
{
while (eof==0)      /* do until no more input */
{
if(amatch("char",4)){declglb(cchar);ns();}
else if(amatch("int",3)){declglb(cint);ns();}
else if(match("#asm"))doasm();
else if(match("#include"))doinclude();
else if(match("#define"))addmac();
else newfunc();
blanks();   /* force eof if pending */
}
}
/* ### cpcn-3 */
dumpublics()
{
}
/*                  */
/*  Dump the literal pool       */
/*                  */
dumplits()
{int j,k;
if (litptr==0) return;  /* if nothing there, exit...*/
printlabel(litlab); /* print literal label */
k=0;            /* init an index... */
while (k<litptr)    /*  to loop with */
{defbyte(); /* pseudo-op to define byte */
j=10;       /* max bytes per line */
while(j--)
{outdec((litq[k++]&127));
if ((j==0) | (k>=litptr))
{nl();      /* need <cr> */
break;
}
outbyte(' ');   /* separate bytes */
}
}
ol("RS 0");
litptr=0;
litlab=getlabel();
}
/*  print page zero variables */
zpage()
{
outstr(" SET ");
outdec(page);
nl();
page = page -2;
}
/*                  */
/*  Dump all static variables   */
/*                  */
/*                  */
/*  Dump single static variables   */
/*                  */
dumpgbl(cptr)char * cptr;
{
int j;
outname(cptr);tab();
if(cptr[ident]!= array) zpage();
else {
nl();
/* output name as label... */
j=((cptr[offset]&BMASK)+
((cptr[offset+1]&BMASK)<<BITWRD));
/* calc # bytes */
intstorage(j,cptr[type]==cchar);
}
}
/*                  */
/*  Report errors for user      */
/*                  */
errorsummary()
{
nl();
outstr("There were ");
outdec(errcnt); /* total # errors */
outstr(" errors in compilation.");
nl();
}
/* ### cpcn-5 */
/*  Greet User  */
hello()
{
/* clrscreen(); * clear screen function */
nl();      /* print banner */
pl(VERSION);
} /* end of hello */
yesno(s) char *s;
{
pl(s);
kill();
gets(line);
if(ch()=='Y') return 1;
if(ch()=='y') return 1;
return 0;
}
see()
{
/* see if user wants to be sure to see all errors */
errstop=yesno("stop on errors y/n?");
ctext=yesno("text to appear y/n?");
nxtlab=50;
}
/*                  */
/*  Get output filename     */
/*                  */
openout()
{
output=0;       /* start with none */
while(output==0)
{
kill();
pl("Output filename? "); /* ask...*/
gets(line); /* get a filename */
if(ch()==0)break;   /* none given... */
if((output=fopen(line,"w"))==NULL) /* if given, open */
{output=0;  /* can't open */
error("Open failure!");
}
}
kill();         /* erase line * /
*/
}
/*                  */
/*  Get (next) input file       */
/*                  */
openin()
{
input=0;        /* none to start with */
while(input==0){    /* any above 1 allowed */
kill();     /* clear line */
pl("Input filename? ");
gets(line); /* get a name */
if(ch()==0) break;
if((input=fopen(line,"r"))!=NULL)
newfile();          /* gtf 7/16/80 */
else {  input=0;    /* can't open it */
pl("Open failure");
}
}
kill();     /* erase line */
}
/*                  */
/*  Reset line count, etc.      */
/*          gtf 7/16/80 */
newfile()
{
lineno  = 0;    /* no lines read */
fnstart = 0;    /* no fn. start yet. */
currfn  = NULL; /* because no fn. yet */
infunc  = 0;    /* therefore not in fn. */
/* end newfile */
}
/* ### cpcn-6 */
/*                  */
/*  Open an include file        */
/*                  */
doinclude()
{
blanks();   /* skip over to name */
toconsole();                    /* gtf 7/16/80 */
outstr("#include "); outstr(line+lptr); nl();
tofile();
if(input2)                  /* gtf 7/16/80 */
error("Cannot nest include files");
else if((input2=fopen(line+lptr,"r"))==NULL)
{input2=0;
error("Open failure on include file");
}
else {  saveline = lineno;
savecurr = currfn;
saveinfn = infunc;
savestart= fnstart;
newfile();
}
kill();     /* clear rest of line */
/* so next read will come from */
/* new file (if open */
}
/*                  */
/*  Close an include file       */
/*          gtf 7/16/80 */
endinclude()
{
toconsole();
outstr("#end include"); nl();
tofile();
input2  = 0;
lineno  = saveline;
currfn  = savecurr;
infunc  = saveinfn;
fnstart = savestart;
/* end endinclude */}
/*                  */
/*  Close the output file       */
/*                  */
closeout()
{
tofile();   /* if diverted, return to file */
if(output)fclose(output); /* if open, close it */
output=0;       /* mark as closed */
}
/* ### cpcn-7 */
/*                  */
/* initialize global objects simple constants */
intstorage(dim,size) int dim,size; {
int d;
d = dim;
if(match("=")) {  /* int constant values */
match("{");
while(dim) {
defstorage(size,1,0);dim=dim-1;
match(",");
}
needbrack("}");
}
/* int zero values */
if(dim) defstorage(size,0,dim);
if(size)defstorage(size,0,0);
}
/*  Declare a static variable   */
/*    (i.e. define for use)     */
/*                  */
/* makes an entry in the symbol table so subsequent */
/*  references can call symbol by name  */
/*  program header has jmp to _main so this is skipped over */
declglb(typ)        /* typ is cchar or cint */
int typ;
{   int k,j,g;char sname[namesiz];
while(1)
{while(1)
{if(endst())return; /* do line */
k=1;        /* assume 1 element */
if(match("*"))  /* pointer ? */
j=pointer;  /* yes */
else j=variable; /* no */
if (symname(sname)==0) /* name ok? */
illname(); /* no... */
if(findglb(sname)) /* already there? */
multidef(sname);
if (match("["))     /* array? */
{k=needsub();   /* get size */
if(k)j=array;   /* !0=array */
else j=pointer; /* 0=ptr */
}
g=addglb(sname,j,typ,k); /* add symbol */
dumpgbl(g); /* emit values */
break;
}
if (match(",")==0) return; /* more? */
}
}
/*                  */
/*  Declare local variables     */
/*  (i.e. define for use)       */
/*                  */
/* works just like "declglb" but modifies machine stack */
/*  and adds symbol table entry with appropriate */
/*  stack offset to find it again           */
declloc(typ)        /* typ is cchar or cint */
int typ;
{
int k,j; char sname[namesiz];
while(1)
{while(1)
{if(endst())return;
if(match("*"))
j=pointer;
else j=variable;
if (symname(sname)==0)
illname();
if(findloc(sname))
multidef(sname);
if (match("["))
{k=needsub();
if(k)
{j=array;
if(typ==cint)k=k+k;
else if(k&1) k=k+1;
}
else
{j=pointer;
k=2;
}
}
else
if((typ==cchar)
&(j!=pointer))
k=2;else k=2;
/* change machine stack */
Zsp=modstk(Zsp-k);
addloc(sname,j,typ,Zsp);
break;
}
if (match(",")==0) return;
}
}
/* ### cpcn-8 */
/*  >>>>>> start of cc2 <<<<<<<<    */
/*                  */
/*  Get required array size     */
/*                  */
/* invoked when declared variable is followed by "[" */
/*  this routine makes subscript the absolute */
/*  size of the array. */
needsub()
{
int num[1];
if(match("]"))return 0; /* null size */
if (number(num)==0) /* go after a number */
{error("must be constant"); /* it isn't */
num[0]=1;       /* so force one */
}
if (num[0]<0)
{error("negative size illegal");
num[0]=(-num[0]);
}
needbrack("]");     /* force single dimension */
return num[0];      /* and return size */
}
/*                  */
/* ### cpcn-9 */
/*  Begin a function        */
/*                  */
/* Called from "parse" this routine tries to make a function */
/*  out of what follows.    */
newfunc()
{
char n[namesiz]; /* ptr => currfn,  gtf 7/16/80 */
int loc;
if (symname(n)==0)
{
if(eof==0) error("illegal function or declaration");
kill(); /* invalidate line */
return;
}
fnstart=lineno;     /* remember where fn began  gtf 7/2/80 */
infunc=1;       /* note, in function now.   gtf 7/16/80 */
if(currfn=findglb(n))   /* already in symbol table ? */
{if(currfn[ident]!=function)multidef(n);
/* already variable by that name */
else if(currfn[offset]==function)multidef(n);
/* already function by that name */
else currfn[offset]=function;
/* otherwise we have what was earlier*/
/*  assumed to be a function */
}
/* if not in table, define as a function now */
else currfn=addglb(n,function,cint,function);
toconsole();                    /* gtf 7/16/80 */
outstr("====== "); outstr(n); outstr("()"); nl();
tofile();
/* we had better see open paren for args... */
if(match("(")==0)error("missing open paren");
outname(n); nl();   /* print function name */
argstk=0;       /* init arg count */
while(match(")")==0)    /* then count args */
/* any legal name bumps arg count */
{if(symname(n))argstk=argstk+2;
else{error("illegal argument name");junk();}
blanks();
/* if not closing paren, should be comma */
if(streq(line+lptr,")")==0)
{if(match(",")==0)
error("expected comma");
}
if(endst())break;
}
locptr=startloc;    /* "clear" local symbol table*/
Zsp=0;
loc=argstk;
argstk=0;           /* preset stack ptr */
while(argstk<loc)
/* now let user declare what types of things */
/*  those arguments were */
{
if(amatch("char",4)){getarg(cchar,loc);ns();}
else if(amatch("int",3)){getarg(cint,loc);ns();}
else{error("wrong number args");break;}
}
Zsp=0;
modstk(-16); /* enter routine */
if(statement()!=streturn) /* do a statement, but if */
/* it's a return, skip */
/* cleaning up the stack */
{modstk(16);
zret();
}
Zsp=0;          /* reset stack ptr again */
locptr=startloc;    /* deallocate all locals */
infunc=0;       /* not in fn. any more      gtf 7/2/80 */
dumplits();     /* 7/13/19 */
}
/* ### cpcn-10 */
/*                  */
/*  Declare argument types      */
/*                  */
/* called from "newfunc" this routine adds an entry in the */
/*  local symbol table for each named argument */
getarg(t,l)     /* t = cchar or cint  l local */
int t,l;
{
char n[namesiz];int j;
while(1)
{
argstk=argstk+2;
if(argstk>l)return; /* no more args */
if(match("*"))j=pointer;
else j=variable;
if(symname(n)==0) illname();
if(findloc(n))multidef(n);
if(match("["))  /* pointer ? */
/* it is a pointer, so skip all */
/* stuff between "[]" */
{while(inbyte()!=']')
if(endst())break;
j=pointer;
/* add entry as pointer */
}
addloc(n,j,t,argstk);
if(endst())return;
if(match(",")==0)error("expected comma");
}
}
/* ### cpcn-11 */
/*  Statement parser        */
/*                  */
/* called whenever syntax requires  */
/*  a statement.             */
/*  this routine performs that statement */
/*  and returns a number telling which one */
statement()
{
/* comment out ctrl-C check since ctrl-break will work on PC */
/*  if(cpm(11,0) & 1) */    /* check for ctrl-C     gtf 7/17/80 */
/*  if(getchar()==3) */
/*  aborting(); */
Lsp =0;
if ((ch()==0) & (eof)) return;
else if(amatch("char",4))
{declloc(cchar);ns();}
else if(amatch("int",3))
{declloc(cint);ns();}
else if(match("{"))compound();
else if(amatch("if",2))
{doif();lastst=stif;}
else if(amatch("while",5))
{dowhile();lastst=stwhile;}
else if(amatch("return",6))
{doreturn();ns();lastst=streturn;}
else if (amatch("switch",6))
{doswitch();lastst=stswitch;}
else if(amatch("break",5))
{dobreak();ns();lastst=stbreak;}
else if(amatch("continue",8))
{docont();ns();lastst=stcont;}
else if(match(";"));
else if(match("#asm"))
{doasm();lastst=stasm;}
/* if nothing else, assume it's an expression */
else{expression(0);ns();lastst=stexp;}
return lastst;
}
/*                  */
/* ### cpcn-12 */
/*  Semicolon enforcer      */
/*                  */
/* called whenever syntax requires a semicolon */
ns()    {if(match(";")==0)error("missing semicolon");}
/*                  */
/*  Compound statement      */
/*                  */
/* allow any number of statements to fall between "{}" */
compound()
{
++ncmp;     /* new level open */
while (match("}")==0) statement(); /* do one */
--ncmp;     /* close current level */
}
/*                  */
/*      "if" statement      */
/*                  */
doif()
{
int flev,fsp,flab1,flab2;
flev=locptr;    /* record current local level */
fsp=Zsp;        /* record current stk ptr */
flab1=getlabel(); /* get label for false branch */
test(flab1);    /* get expression, and branch false */
statement();    /* if true, do a statement */
Zsp=modstk(fsp);    /* then clean up the stack */
locptr=flev;    /* and deallocate any locals */
if (amatch("else",4)==0)    /* if...else ? */
/* simple "if"...print false label */
{printlabel(flab1);
return;     /* and exit */
}
/* an "if...else" statement. */
jump(flab2=getlabel()); /* jump around false code */
printlabel(flab1);    /* print false label */
statement();        /* and do "else" clause */
Zsp=modstk(fsp);        /* then clean up stk ptr */
locptr=flev;        /* and deallocate locals */
printlabel(flab2);     /* print true label */
}
/*                  */
/*  "while" statement       */
/*                  */
dowhile()
{
int wq[wqsiz];      /* allocate local queue */
wq[wqsym]=locptr;   /* record local level */
wq[wqsp]=Zsp;       /* and stk ptr */
wq[wqloop]=getlabel();  /* and looping label */
wq[wqlab]=getlabel();   /* and exit label */
addwhile(wq);       /* add entry to queue */
/* (for "break" statement) */
printlabel(wq[wqloop]); /* loop label */
test(wq[wqlab]);    /* see if true */
statement();        /* if so, do a statement */
Zsp = modstk(wq[wqsp]); /* zap local vars: 9/25/80 gtf */
jump(wq[wqloop]);   /* loop to label */
printlabel(wq[wqlab]); /* exit label */
locptr=wq[wqsym];   /* deallocate locals */
delwhile();     /* delete queue entry */
}
/*                  */
/* ### cpcn-13 */
/*                  */
/*  "return" statement      */
/*                  */
doreturn()
{
/* if not end of statement, get an expression */
if(endst()==0)expression(0);
modstk(16); /* clean up stk */
zret();     /* and exit function */
}
/*                  */
/*  "break" statement       */
/*                  */
dobreak()
{
int *ptr;
/* see if any "whiles" are open */
if ((ptr=readwhile())==0) return;   /* no */
modstk((ptr[wqsp]));    /* else clean up stk ptr */
jump(ptr[wqlab]);   /* jump to exit label */
}
/*                  */
/*  "continue" statement        */
/*                  */
docont()
{
int *ptr;
/* see if any "whiles" are open */
if ((ptr=readwhile())==0) return;   /* no */
modstk((ptr[wqsp]));    /* else clean up stk ptr */
jump(ptr[wqloop]);  /* jump to loop label */
}
/*                  */
/*  "asm" pseudo-statement      */
/*                  */
/* enters mode where assembly language statement are */
/*  passed intact through parser    */
doasm()
{
cmode=0;        /* mark mode as "asm" */
while (1)
{inline();  /* get and print lines */
if (match("#endasm")) break;    /* until... */
if(eof)break;
outstr(line);
nl();
}
kill();     /* invalidate line */
cmode=1;        /* then back to parse level */
}
/* ### cpcn-14 */
/*  >>>>> start of cc3 <<<<<<<<<    */
/*                  */
/*  Perform a function call     */
/*                  */
/* called from heir11, this routine will either call */
/*  the named function, or if the supplied ptr is */
/*  zero, will call s+2     */
callfunction(ptr)
char *ptr;  /* symbol table entry (or 0) */
{   int nargs;
nargs=0;
blanks();   /* already saw open paren */
if(ptr==0)zpush();  /* calling BX */
while(streq(line+lptr,")")==0)
{if(endst())break;
expression(0);  /* get an argument */
nargs=nargs+2;  /* count args*2 */
apush(nargs);   /* push argument */
if (match(",")==0) break;
}
needbrack(")");
if(ptr!=0)zcall(ptr);
else callstk();
Zsp=modstk(Zsp);    /* clean up arguments */
}
junk()
{   if(an(inbyte()))
while(an(ch()))gch();
else while(an(ch())==0)
{if(ch()==0)break;
gch();
}
blanks();
}
endst()
{   blanks();
return ((streq(line+lptr,";")|(ch()==0)));
}
illname()
{   error("illegal symbol name");junk();}
multidef(sname)
char *sname;
{   error("already defined");
comment();
outstr(sname);nl();
}
needbrack(str)
char *str;
{   if (match(str)==0)
{error("missing bracket");
comment();outstr(str);nl();
}
}
needlval()
{   error("must be lvalue");
}
/* ### cpcn-15 */
findglb(sname)
char *sname;
{   char *ptr;
ptr=startglb;
while(ptr!=glbptr)
{if(astreq(sname,ptr,namemax))return ptr;
ptr=ptr+symsiz;
}
return 0;
}
findloc(sname)
char *sname;
{   char *ptr;
ptr=startloc;
while(ptr!=locptr)
{if(astreq(sname,ptr,namemax))return ptr;
ptr=ptr+symsiz;
}
return 0;
}
addglb(sname,id,typ,value)
char *sname,id,typ;
int value;
{   char *ptr;
if(cptr=findglb(sname))return cptr;
if(glbptr>=endglb)
{error("global symbol table overflow");
return 0;
}
cptr=ptr=glbptr;
while(an(*ptr++ = *sname++));   /* copy name */
cptr[ident]=id;
cptr[type]=typ;
cptr[storage]=statik;
cptr[offset]=value;
cptr[offset+1]=value>>BITWRD;
glbptr=glbptr+symsiz;
return cptr;
}
addloc(sname,id,typ,value)
char *sname,id,typ;
int value;
{   char *ptr;
if(cptr=findloc(sname))return cptr;
if(locptr>=endloc)
{error("local symbol table overflow");
return 0;
}
cptr=ptr=locptr;
while(an(*ptr++ = *sname++));   /* copy name */
cptr[ident]=id;
cptr[type]=typ;
cptr[storage]=stkloc;
cptr[offset]=value;
cptr[offset+1]=value>>BITWRD;
locptr=locptr+symsiz;
return cptr;
}
/* ### cpcn-16 */
/* Test if next input string is legal symbol name */
symname(sname)
char *sname;
{   int k;
blanks();
if(alpha(ch())==0)return 0;
k=0;
while(an(ch()))sname[k++]=gch();
sname[k]=0;
return 1;
}
/* ### cpcn-17 */
/* Return next avail internal label number */
getlabel()
{   return(++nxtlab);
}
/* Print specified number as label */
printlabel(label)
int label;
{   outasm("?");
outdec(label);
nl();
}
/* Test if given character is alpha */
alpha(c)
char c;
{   c=c&127;
return(((c>='a')&(c<='z'))|
((c>='A')&(c<='Z'))|
(c=='_'));
}
/* Test if given character is numeric */
numeric(c)
char c;
{   c=c&127;
return((c>='0')&(c<='9'));
}
/* Test if given character is alphanumeric */
an(c)
char c;
{   return((alpha(c))|(numeric(c)));
}
/* Print a carriage return and a string only to console */
pl(str)
char *str;
{   int k;
k=0;
putchar(eol);
while(str[k])putchar(str[k++]);
}
addwhile(ptr)
int ptr[];
{
int k;
if (wqptr==wqmax)
{error("too many active whiles");return;}
k=0;
while (k<wqsiz)
{*wqptr++ = ptr[k++];}
}
delwhile()
{if(readwhile()) wqptr=wqptr-wqsiz;
}
readwhile()
{
if (wqptr==wq){error("no active whiles");return 0;}
else return (wqptr-wqsiz);
}
ch()
{   return(line[lptr]&127);
}
nch()
{   if(ch()==0)return 0;
else return(line[lptr+1]&127);
}
gch()
{   if(ch()==0)return 0;
else return(line[lptr++]&127);
}
kill()
{   lptr=0;
line[lptr]=0;
}
/* ### cpcn-18 */
inbyte()
{
while(ch()==0)
{if (eof) return 0;
inline();
preprocess();
}
return gch();
}
inchar()
{
if(ch()==0)inline();
if(eof)return 0;
return(gch());
}
inline()
{
int k,unit;
while(1)
{if (input==0)  {eof=1;return;}
if((unit=input2)==0)unit=input;
kill();
while((k=getc(unit))>0)
{if((k==eol)|(lptr>=linemax))break;
 line[lptr++]=k;
}
line[lptr]=0;   /* append null */
lineno++;   /* read one more line       gtf 7/2/80 */
if(k<=0)
{fclose(unit);
if(input2)endinclude();     /* gtf 7/16/80 */
else input=0;
}
if(lptr)
{if((ctext)&(cmode))
{comment();
outstr(line);
nl();
}
lptr=0;
return;
}
}
}
/* ### cpcn-19 */
/*  >>>>>> start of cc4 <<<<<<< */
preprocess()
{   int k;
char c,sname[namesiz];
if(cmode==0)return;
mptr=lptr=0;
while(ch())
{if((ch()==' ')|(ch()==9))
predel();
else if(ch()=='"')
prequote();
else if(ch()==39)
preapos();
else if((ch()=='/')&(nch()=='*'))
precomm();
else if(alpha(ch()))    /* from an(): 9/22/80 gtf */
{k=0;
while(an(ch()))
{if(k<namemax)sname[k++]=ch();
gch();
}
sname[k]=0;
if(k=findmac(sname))
while(c=macq[k++])
keepch(c);
else
{k=0;
while(c=sname[k++])
keepch(c);
}
}
else keepch(gch());
}
keepch(0);
if(mptr>=mpmax)error("line too long");
lptr=mptr=0;
while(line[lptr++]=mline[mptr++]);
lptr=0;
}
/* ### cpcn-20 */
keepch(c)
char c;
{   mline[mptr]=c;
if(mptr<mpmax)mptr++;
return c;
}
predel()
{keepch(' ');
while((ch()==' ')|
(ch()==9))
gch();
}
prequote()
{keepch(ch());
gch();
while(ch()!='"')
{if(ch()==0)
{error("missing quote");
break;
}
keepch(gch());
}
gch();
keepch('"');
}
preapos()
{keepch(39);
gch();
while(ch()!=39)
{if(ch()==0)
{error("missing apostrophe");
break;
}
keepch(gch());
}
gch();
keepch(39);
}
precomm()
{inchar();inchar();
while(((ch()=='*')&
(nch()=='/'))==0)
{if(ch()==0)inline();
else inchar();
if(eof)break;
}
inchar();inchar();
}
/* ### cpcn-21 */
addmac()
{   char sname[namesiz];
int k;
if(symname(sname)==0)
{illname();
kill();
return;
}
k=0;
while(putmac(sname[k++]));
while(ch()==' ' | ch()==9) gch();
while(putmac(gch()));
if(macptr>=macmax)error("macro table full");
}
putmac(c)
char c;
{   macq[macptr]=c;
if(macptr<macmax)macptr++;
return c;
}
findmac(sname)
char *sname;
{   int k;
k=0;
while(k<macptr)
{if(astreq(sname,macq+k,namemax))
{while(macq[k++]);
return k;
}
while(macq[k++]);
while(macq[k++]);
}
return 0;
}
/* ### cpcn-22 */
/* direct output to console     gtf 7/16/80 */
toconsole()
{
saveout = output;
output = 0;
/* end toconsole */}
/* direct output back to file       gtf 7/16/80 */
tofile()
{
if(saveout)
output = saveout;
saveout = 0;
/* end tofile */}
outbyte(c)
char c;
{
if(c==0)return 0;
if(output)
{if((putc(c,output))<=0)
{closeout();
error("Output file error");
aborting();         /* gtf 7/17/80 */
}
}
else putchar(c);
return c;
}
outstr(ptr)
char ptr[];
{
int k;
k=0;
while(outbyte(ptr[k++]));
}
/* write text destined for the assembler to read */
/* (i.e. stuff not in comments)         */
/*  gtf  6/26/80 */
/* @ char prints Acc value 0 A 1 X */
outasm(ptr)
char *ptr;
{
int ch;
while(ch=*ptr++)
{ if(ch=='@')
{ ch='A';if(Acc==1) ch='X';}
else ch=raise(ch);
outbyte(ch);
}
/* end outasm */}
nl()
{outbyte(eol);}
tab()
{outbyte(9);}
/* ### cpcn-23 */
error(ptr)
char ptr[];
{   int k;
char junk[81];
toconsole();
/*  bell();  */
outstr("Line "); outdec(lineno); outstr(", ");
if(infunc==0)
outbyte('(');
if(currfn==NULL)
outstr("start of file");
else    outstr(currfn+name);
if(infunc==0)
outbyte(')');
outstr(" + ");
outdec(lineno-fnstart);
outstr(": ");  outstr(ptr);  nl();
outstr(line); nl();
k=0;    /* skip to error position */
while(k<lptr){
if(line[k++]==9)
tab();
else    outbyte(' ');
}
outbyte('^');  nl();
++errcnt;
if(errstop){
pl("Continue (Y,n,g) ? ");
gets(junk);
k=junk[0];
if((k=='N') | (k=='n'))
aborting();
if((k=='G') | (k=='g'))
errstop=0;
}
tofile();
/* end error */}
ol(ptr)
char ptr[];
{
ot(ptr);
nl();
}
ot(ptr)
char ptr[];
{
tab();
outasm(ptr);
}
/* ### cpcn-24 */
streq(str1,str2)
char str1[],str2[];
{
int k;
k=0;
while (str2[k])
{if ((str1[k])!=(str2[k])) return 0;
k++;
}
return k;
}
astreq(str1,str2,len)
char str1[],str2[];int len;
{
int k;
k=0;
while (k<len)
{if ((str1[k])!=(str2[k]))break;
if(str1[k]==0)break;
if(str2[k]==0)break;
k++;
}
if (an(str1[k]))return 0;
if (an(str2[k]))return 0;
return k;
}
match(lit)
char *lit;
{
int k;
blanks();
if (k=streq(line+lptr,lit))
{lptr=lptr+k;
return 1;
}
return 0;
}
amatch(lit,len)
char *lit;int len;
{
int k;
blanks();
if (k=astreq(line+lptr,lit,len))
{lptr=lptr+k;
while(an(ch())) inbyte();
return 1;
}
return 0;
}
/* ### cpcn-25 */
blanks()
{while(1)
{while(ch()==0)
{inline();
preprocess();
if(eof)break;
}
if(ch()==' ')gch();
else if(ch()==9)gch();
else return;
}
}
outdec(number)
int number;
{
int k,zs;
char c;
zs = 0;
k=KDIV;
if (number<0)
{number=(-number);
outbyte('-');
}
if(number<1000) k=100;
while (k>=1)
{
c=number/k + '0';
if ((c!='0')|(k==1)|(zs))
{zs=1;outbyte(c);}
number=number%k;
k=k/10;
}
}
/* return the length of a string */
/* gtf 4/8/80 *
strlen(s)
char *s;
{   char *t;
t = s;
while(*s) s++;
return(s-t);
/* end strlen }     */
/* convert lower case to upper */
/* gtf 6/26/80 */
raise(c)
char c;
{
if((c>='a') & (c<='z'))
c = c - 'a' + 'A';
return(c);
/* end raise */}
/* ### cpcn-26 */
/*  >>>>>>> start of cc5 <<<<<<<    */
/* modified 9/25/80: putstk() */
/* 0 use acc 1 use index reg  */
expression(i) int i;
{
int lval[2],t;
t=Acc;
Acc=i;
if(heir1(lval))rvalue(lval);
Acc=t;
}
heir1(lval)
int lval[];
{
int k,lval2[2],t;
k=heir2(lval);
if (match("="))
{if(k==0){needlval();return 0;}
if (lval[1]){t=Acc;Acc=1;zpush();Acc=t;}
if(heir1(lval2))rvalue(lval2);
store(lval);
return 0;
}
else return k;
}
heir2(lval)
int lval[];
{   int k,lval2[2];
k=heir3(lval);
blanks();
if(ch()!='|')return k;
if(k)k=ivalue(lval);else k=1;
while(1)
{if (match("|"))
{if(k)zpush();
if(heir3(lval2)) rvalue(lval2);
if(k)zor();
else {getefa("OR",*lval);k=1;}
}
else return 0;
}
}
/* ### cpcn-27 */
heir3(lval)
int lval[];
{   int k,lval2[2];
k=heir4(lval);
blanks();
if(ch()!='^')return k;
if(k)k=ivalue(lval);else k=1;
while(1)
{if (match("^"))
{if(k)zpush();
if(heir4(lval2))rvalue(lval2);
if(k){zxor();}
else {getefa("XOR",*lval);k=1;}
}
else return 0;
}
}
heir4(lval)
int lval[];
{   int k,lval2[2];
k=heir5(lval);
blanks();
if(ch()!='&')return k;
if(k)k=ivalue(lval); else k=1;
while(1)
{if (match("&"))
{if(k)zpush();
if(heir5(lval2))rvalue(lval2);
if(k){zand();}
else{getefa("AND",*lval);k=1;}
}
else return 0;
}
}
heir5(lval)
int lval[];
{
int k,lval2[2];
k=heir6(lval);
blanks();
if((streq(line+lptr,"==")==0)&
(streq(line+lptr,"!=")==0))return k;
if(k)k=ivalue(lval);else k=1;
while(1)
{if (match("=="))
{if(k)zpush();
if(heir6(lval2))rvalue(lval2);
if(k){zsub();}
else getefa("CAD",*lval);
zeq();
k=1;
}
else if (match("!="))
{
if(k)zpush();
if(heir6(lval2))rvalue(lval2);
if(k){zsub();}
else getefa("CAD",*lval);
k=1;
zne();
}
else return 0;
}
}
/* ### cpcn-28 */
heir6(lval)
int lval[];
{
int k;
k=heir7(lval);
blanks();
if((streq(line+lptr,"<")==0)&
(streq(line+lptr,">")==0)&
(streq(line+lptr,"<=")==0)&
(streq(line+lptr,">=")==0))return k;
if(streq(line+lptr,">>"))return k;
if(streq(line+lptr,"<<"))return k;
if(k)k=ivalue(lval);else k=1;
while(1)
{if (match("<="))
{if(heir6wrk(k,1,lval))zle();k=1;
}
else if (match(">="))
{if(heir6wrk(k,2,lval))zge();k=1;
}
else if((streq(line+lptr,"<"))&
(streq(line+lptr,"<<")==0))
{inbyte();
if(heir6wrk(k,3,lval))zlt();k=1;
}
else if((streq(line+lptr,">"))&
(streq(line+lptr,">>")==0))
{inbyte();
if(heir6wrk(k,4,lval))zgt();k=1;
}
else return 0;
}
}
/* ### cpcn-29 */
heir6wrk(i,k,lval)
int i,k,lval[];
{
int lval2[2];
if(i)zpush();
if(heir7(lval2))rvalue(lval2);
if(i){zsub();}
else getefa("CAD",*lval);
if(cptr=lval[0])
if(cptr[ident]==pointer)
{heir6op(k);
return 0;
}
if(cptr=lval2[0])
if(cptr[ident]==pointer)
{heir6op(k);
return 0;
}
return 1;
}
heir6op(k)
int k;
{
if(k==1) ule();
else if(k==2) uge();
else if(k==3) ult();
else ugt();
}
/* ### cpcn-30 */
/*  >>>>>> start of cc6 <<<<<<  */
heir7(lval)
int lval[];
{
int k,lval2[2];
k=heir8(lval);
blanks();
if((streq(line+lptr,">>")==0)&
(streq(line+lptr,"<<")==0))return k;
if(k)rvalue(lval);
while(1)
{if (match(">>"))
{zpush();
if(heir8(lval2))rvalue(lval2);
asr();
}
else if (match("<<"))
{zpush();
if(heir8(lval2))rvalue(lval2);
asl();
}
else return 0;
}
}
/* ### cpcn-31 */
heir8(lval)
int lval[];
{
int k,lval2[2];
k=heir9(lval);
blanks();
if((ch()!='+')&(ch()!='-'))return k;
if(k)k=ivalue(lval);else k=1;
while(1)
{if (match("+"))
{if(k)zpush();
if(heir9(lval2))rvalue(lval2);
if(cptr=lval[0])
if((cptr[ident]==pointer)&
(cptr[type]==cint))
doublereg();
if(k){zadd();}
else {getefa("ADD",*lval);k=1;}
}
else if (match("-"))
{if(k)zpush();
if(heir9(lval2))rvalue(lval2);
if(cptr=lval[0])
if((cptr[ident]==pointer)&
(cptr[type]==cint))
doublereg();
if(k){zsub();}
else {getefa("CAD",*lval);k=1;}
}
else return 0;
}
}
/* ### cpcn-32 */
heir9(lval)
int lval[];
{
int k,lval2[2];
k=heir10(lval);
blanks();
if((ch()!='*')&(ch()!='/')&
(ch()!='%'))return k;
if(k)rvalue(lval);
while(1)
{if (match("*"))
{zpush();
if(heir9(lval2))rvalue(lval2);
mult();
}
else if (match("/"))
{zpush();
if(heir10(lval2))rvalue(lval2);
div();
}
else if (match("%"))
{zpush();
if(heir10(lval2))rvalue(lval2);
zmod();
}
else return 0;
}
}
/* ### cpcn-33 */
heir10(lval)
int lval[];
{
int k;
if(match("++"))
{if((k=heir10(lval))==0)
{needlval();
return 0;
}
heir10inc(lval,1);
return 0;
}
else if(match("--"))
{if((k=heir10(lval))==0)
{needlval();
return 0;
}
heir10inc(lval,-1);
return 0;
}
else if (match("-"))
{k=heir10(lval);
if (k) rvalue(lval);
neg();
return 0;
}
else if(match("*"))
{heir10as(lval);
return 1;
}
else if(match("&"))
{k=heir10(lval);
if(k==0)
{error("illegal address");
return 0;
}
else if(lval[1])return 0;
else
{heir10at(lval);
return 0;
}
}
else
{k=heir11(lval);
if(match("++"))
{if(k==0)
{needlval();
return 0;
}
heir10id(lval,1);
return 0;
}
else if(match("--"))
{if(k==0)
{needlval();
return 0;
}
heir10id(lval,-1);
return 0;
}
else return k;
}
}
/* ### cpcn-34 */
heir10inc(lval,i)
int lval[],i;
{
char *ptr;
if(lval[1])zpush();
rvalue(lval);
ptr=lval[0];
if((ptr[ident]==pointer)&
(ptr[type]==cint))
inc(i+i);
else inc(i);
store(lval);
}
heir10as(lval)
int lval[];
{
int k,t;
char *ptr;
t=Acc;
Acc=1;
k=heir10(lval);
if(k)rvalue(lval);
lval[1]=cint;
if(ptr=lval[0])lval[1]=ptr[type];
lval[0]=0;
Acc=t;
}
/* ### cpcn-35 */
heir10at(lval)
int lval[];
{
char *ptr;
getloc(lval[0]);
lval[1]=ptr[type];
}
heir10id(lval,i)
int lval[],i;
{
char *ptr;
if(lval[1])zpush();
rvalue(lval);
ptr=lval[0];
if((ptr[ident]==pointer)&
(ptr[type]==cint))
i=i+i;
inc(i);
store(lval);
inc(-i);
}
/* ### cpcn-36 */
/*  >>>>>> start of cc7 <<<<<<  */
heir11(lval)
int *lval;
{   int k,t;char *ptr;
k=primary(lval);
ptr=lval[0];
blanks();
if((ch()=='[')|(ch()=='('))
while(1)
{
if(match("["))
{ k=1;t=Acc;
if(ptr==0)
{error("can't subscript");
junk();
needbrack("]");
return 0;
}
else if(ptr[ident]==pointer)k=0;
else if(ptr[ident]!=array)
{error("can't subscript");
k=0;
}
if(k)zpush();
expression(1);
needbrack("]");
Acc=1;
if(ptr[type]==cint)doublereg();
if(k){zadd();}
else {getefa("ADD",*lval);}
Acc=t;
lval[1]=ptr[type];
k=1;
}
else if(match("("))
{if(ptr==0)
{callfunction(0);
}
else if(ptr[ident]!=function)
{rvalue(lval);
callfunction(0);
}
else callfunction(ptr);
k=lval[0]=0;
}
else return k;
}
if(ptr==0)return k;
if(ptr[ident]==function)
{immed();
outname(ptr);
nl();
return 0;
}
return k;
}
/* ### cpcn-37 */
primary(lval)
int *lval;
{   char *ptr,sname[namesiz];int num[1];
int k;
if(match("("))
{k=heir1(lval);
needbrack(")");
return k;
}
if(symname(sname))
{if(ptr=findloc(sname))
{
lval[0]=ptr;
lval[1]=0;
if(ptr[ident]!=array)return 1;
getloc(ptr);
lval[1]=ptr[type];
return 0;
}
if(ptr=findglb(sname))
if(ptr[ident]!=function)
{lval[0]=ptr;
lval[1]=0;
if(ptr[ident]!=array)return 1;
getloc(ptr);
lval[1]=ptr[type];
return 0;
}
ptr=addglb(sname,function,cint,0);
lval[0]=ptr;
lval[1]=0;
return 0;
}
if(constant(num))
return(lval[0]=lval[1]=0);
else
{error("invalid expression");
junk();
return 0;
}
}
/* ### cpcn-38 */
store(lval)
int *lval;
{   if (lval[1]==0)putmem(lval[0]);
else putstk(lval[1]);
}
rvalue(lval)
int *lval;
{   if((lval[0] != 0) & (lval[1] == 0))
getmem(lval[0]);
else indirect(lval[1]);
}
/* true if indirect  */
ivalue(lval)
int *lval;
{   if((lval[0] != 0) & (lval[1] == 0)) return 0;
indirect(lval[1]);
return 1;
}
test(label)
int label;
{
needbrack("(");
expression(0);
needbrack(")");
testjump(label);
}
constant(val)
int val[];
{   int t;
if (number(val))
{
t = val[0];
if((t > -512) & (t< 511)) imms();
else immed();
}
else if (pstr(val))
immed();
else if (qstr(val))
{immed();
outstr("?");
outdec(litlab);
outstr(" + ");}
else return 0;
outdec(val[0]);
nl();
return 1;
}
noconst(val)
int val[];
{
if (number(val)) return 0;
if (pstr(val)) return 0;
return 1;
}
/* ### cpcn-39 */
/* octal digits july 18/19*/
number(val)
int val[];
{   int k,minus;char c;int o;
o=10;
k=minus=1;
while(k)
{k=0;
if (match("+")) k=1;
if (match("-")) {minus=(-minus);k=1;}
}
if(numeric(ch())==0)return 0;
if(ch()== '0') o = 8;
while (numeric(ch()))
{c=inbyte();
k=k*o+(c-'0');
}
if (minus<0) k=(-k);
val[0]=k;
return 1;
}
pstr(val)
int val[];
{   int k;char c;
k=0;
if (match("'")==0) return 0;
while((c=gch())!=39)
k=(k<<BITWRD) + (c&127);
val[0]=k;
return 1;
}
qstr(val)
int val[];
{
char c;
if (match(quote)==0) return 0;
val[0]=litptr;
while (ch()!='"')
{if(ch()==0)break;
if(litptr>=litmax)
{error("string space exhausted");
while(match(quote)==0)
if(gch()==0)break;
return 1;
}
c =gch();
if(c==92) { /* back slash */
/* no EOL check */
c = gch();
if(c=='n') c=eol;
if(c=='b') c =8;
if(c=='t') c =9;
if((c>='0')&(c<='7'))
/* two octal digits hack */
c=((c&7)<<3)+(gch()&7);
}
litq[litptr++]=c;
}
gch();
litq[litptr++]=0;
return 1;
}
/* ### cpcn-40 */
/*  >>>>>> start of cc8 <<<<<<< */
/* Begin a comment line for the assembler */
comment()
{   outbyte('/');
}
/* Put out assembler info before any code is generated */
header()
{
comment();
outstr(VERSION);
nl();
}
/* Print any assembler stuff needed after all code */
trailer()
{
nl();           /* 6 May 80 rj errorsummary() now goes to console */
comment();
outstr(" --- End of Compilation ---");
nl();
ol("END");
}
/* ### cpcn-41 */
/* Print out a name such that it won't annoy the assembler */
/*  (by matching anything reserved, like opcodes.) */
/*  gtf 4/7/80 */
outname(sname)
char *sname;
{   int len, i,j;
outasm("_");
len = strlen(sname);
if(len>(asmpref+asmsuff)){
i = asmpref;
len = len-asmpref-asmsuff;
while(i-- > 0)
outbyte(raise(*sname++));
while(len-- > 0)
sname++;
while(*sname)
outbyte(raise(*sname++));
}
else    outasm(sname);
/* end outname */}
/* generate a simple instruction with the PRIMARY register */
getefa(ops,sym)
char *ops ,*sym;
{
tab();
outstr(ops);
if((sym[ident]!=pointer)&(sym[type]==cchar)) outstr(".");
if(sym[storage]==statik)
{
outasm(" @ Z ");
outname(sym+name);
nl();
}
else
{
outasm(" @ S ");
outdec(((sym[offset]&BMASK)+
((sym[offset+1]&BMASK)<<BITWRD))-
Zsp+16);nl();
}
}
/* Fetch a static memory cell into the primary register */
getmem(sym)
char *sym;
{
getefa("LD",sym);
}
/* Fetch the address of the specified symbol */
/*  into the primary register a*/
getloc(sym)
char *sym;
{
if(sym[storage]==statik)
{
tab();
outasm("LD @ #");
outname(sym+name);
nl();
}
else
{
ot("LEA @ S ");
outdec(((sym[offset]&BMASK)+
((sym[offset+1]&BMASK)<<BITWRD))-
Zsp+16);
nl();
}
}
/* Store the primary register into the specified */
/*  static memory cell */
putmem(sym)
char *sym;
{
getefa("ST",sym);
}
/* Print the instuction using TOS and deciment pointer*/
ztos(s)
char *s;
{
ot(s);
outdec(16-Lsp);   /* skip over return & saved sp */
nl();
Lsp=Lsp-2;
}
/* Store the specified object type in the primary register */
/*  at the address on the top of the stack */
putstk(typeobj)
char typeobj;
{
ztos("LD  X S ");
if(typeobj==cint)
{
ol("ST A X 0");
}
else {
ol("ST. A X 0");
}
}
/* ### cpcn-42 */
/* Fetch the specified object type indirect through the */
/*  primary register into the primary register */
indirect(typeobj)
char typeobj;
{   if(typeobj==cchar) ol("LD. @ X 0");
else ol("LD @ X 0");
}
/* Print partial instruction to get an immediate value */
/*  into the primary register */
immed()
{
ot("LD @ #");
}
/* Print partial instruction to get an short immediate value */
/*  into the primary register */
imms()
{
ot("LD. @ #");
}
/* Push the primary register onto the stack */
zpush()
{
Lsp=Lsp+2;
ot("ST @ S ");outdec(16-Lsp);nl();
}
/* Push the primary register as subroutine argument */
apush(i) int i;
{
ot("ST @ S "); outdec(i);nl();
}
/* Call the specified subroutine name */
zcall(sname)
char *sname;
{   ot("JSR A #");
outname(sname);
nl();
}
/* Call a run-time library routine */
callrts(sname)
char *sname;
{
ot("JSR A #");
outasm(sname);
nl();
/*end callrts*/}
/* Return from subroutine */
zret()
{   ol("JMP A S 0");
}
/* ### cpcn-43 */
/* Perform subroutine call to value on top of stack */
/* use the last free slot as the arg */
callstk()
{     ztos("JSR A S ");  /* fix up -- LEA X S #, JSR A X 0 */
}
/* branch to specified internal label number */
jump(label)
int label;
{   ot("BRA A #");
printlabel(label);
}
/* Test the primary register and jump if false to label */
testjump(label)
int label;
{
ot("BEQ  A #");
printlabel(label);
}
/* Print pseudo-op to define a byte */
defbyte()
{   ot("DB ");
}
/* Print pseudo-op to define a word # or label  */
defwrd(i,l) int i,l;
{
if(l) ot("DW ?"); /* LABEL */
else ot("DW ");   /* CONSTANT */
outdec(i); nl();
}
/*Print pseudo-op to define storage */
defstorage(b,k,v)
int b,k,v;
{
int val[2];
if(k){  /* emit constant */
noconst(val);
v=val[0];
if(b) ot("DB ");
else  ot("DW ");
}
else
{ /* zero fill or pad */
if(b==0) v=v+v;
ot("RS ");
}
outdec(v);nl();
}
/* Modify the stack pointer to the new value indicated */
modstk(newsp)
int newsp;
{  int k;
k=newsp-Zsp;
if(k==0)return newsp;
ot("LEA S S ");
outdec(k);nl();
return newsp;
}
/* ### cpcn-44 */
/* rmult,rdiv,rasl,rasr the only arithmatic runtime calls */
/* Double the primary register */
doublereg()
{   ol("SZL  @ #-1");
}
/* Add the primary and secondary registers */
/*  (results in primary) */
zadd()
{   ztos("ADD @ S ");
}
/* Subtract the primary register from the secondary */
/*  (results in primary) */
zsub()
{
ztos("CAD @ S ");
}
/* Multiply the primary and secondary registers */
/*  (results in primary */
mult()
{
ztos("LD X S ");
callrts("$FF068");
}
/* Divide the secondary register by the primary */
/* (call quotient in secondary, remainder in primary) */
div()
{
ztos("LD X S ");
callrts("$FF06A");
}
/* Compute remainder (mod) of secondary register divided */
/* (call quotient in secondary, remainder in primary) */
zmod()
{
ztos("LD X S ");
callrts("$FF06C");
}
/* Inclusive 'or' the primary and the secondary registers */
/*  (results in primary) */
zor()
{ztos("OR @ S ");}
/* Exclusive 'or' the primary and seconday registers */
/*  (results in primary) */
zxor()
{
ztos("XOR @ S ");}
/* 'And' the primary and secondary registers */
/*  (results in primary) */
zand()
{ztos("AND @ S ");}
/* Arithmetic shift right the secondary register number of */
/*  times in primary (results in primary) */
asr()
{ztos("LD X S ");
callrts("$FF05E");}
/* Arithmetic left shift the secondary register number of */
/*  times in primary (results in primary) */
asl()
{ztos("LD X S ");
callrts("$FF060");}
/* Form two's complement of primary register */
neg()
{ol("CAD. @ #0");}
/* Form one's complement of primary register */
com()
{ol("XOR. @ #-1");}
/* Adjust the primary register by n */
inc(i)  int i;
{ot("ADD. @ #");outdec(i);nl();}
/* Following are the conditional operators */
/* They compare the secondary register against the primary */
/* and put a literal 1 in the primary if the condition is */
/* true, otherwise they clear the primary register */
calltst(s)
char *s;
{
ot(s);outstr(" A #1");nl();
}
/* Test for equal */
zeq()
{calltst("SEQ");}
/* Test for not equal */
zne()
{calltst("SNE");}
/* Test for less than (signed) */
zlt()
{calltst("SLT");}
/* Test for less than or equal to (signed) */
zle()
{calltst("SLE");}
/* Test for greater than (signed) */
zgt()
{calltst("SGT");}
/* Test for greater than or equal to (signed) */
zge()
{calltst("SGE");}
/* Test for less than (unsigned) */
ult()
{calltst("SLO");}
/* Test for less than or equal to (unsigned) */
ule()
{calltst("SLS");}
/* Test for greater than (unsigned) */
ugt()
{calltst("SHI");}
/* Test for greater than or equal to (unsigned) */
uge()
{calltst("SHS");}
/*  <<<<<  End of small-c:PC compiler  >>>>>    */
/*  comple file    rtime.c  */
/*	need opening parn		*/
/*	written	by Mike	Bernson	1/81	*/
needoparn()
{
if (match("("))	return FALSE;
error("Missing left parnthis");
junk();
return TRUE;
}
/*					*/
/*	need closing parn		*/
/*	written	by Mike	Bernson	1/81	*/
needcparn()
{
if (match(")"))	return FALSE;
error("Missing right parnthis");
junk();
return TRUE;
}
/*					*/
/*	need opening brace		*/
/*	written	by Mike	Bernson	1/81	*/
needobrace()
{
if (match("{"))	return FALSE;
error("Missing left brace");
junk();
return TRUE;
}
/*					*/
/*	need closing brace		*/
/*	written	by Mike	Bernson	1/81	*/
needcbrace()
{
if (match("}"))	return FALSE;
error("Missing right brace");
junk();
return TRUE;
}
/* routine to search switch table */
/*	written	by Mike	Bernson	1/81	*/
eswitch(label)
int label;	/* label for switch table */
{
ot(" LD X #");
printlabel(label);
ol(" JMP A #$FF066");
}
/* "Switch" statement		*/
/*					*/
/*	written by Mike Bernson 1/81	*/
/*					*/
doswitch()
{
int value[maxsw];  /* value for case statemant */
int label[maxsw];  /* value for each label */
int count,tenp; 	/* number of switches */
int end_label;  	/* label for default */
int label_switch;	/* used for switch label */
int temp,val[2];
int que[wqsiz];		/* local que area */
count=0;		/* number of case statements */
que[wqsym]=locptr;	/* local vable pointer */
que[wqsp]=Zsp;		/* save current stack pointer */
end_label=		/* default exit label */
que[wqloop]=		/* looping label */
que[wqlab]=		/* loop exit varble */
getlabel();	/* continue label */
addwhile(que);		/* add to while stack */
if (needoparn()) {	/* check to see	if"(" exits */
delwhile();	/* no delete switch entry and */
return;		/* return out of switch	*/
}
expression();		/* expression for switch */
if (needcparn()) {	/* check for ")" */
delwhile();	/* not fould detele que	entry */
return;		/* and exit switch statemant */
}
if (needobrace()) {
delwhile();
return;
}
jump(label_switch=getlabel());
while(1) {
if (amatch("case",4)) {
if (noconst(val)) {
error("Bad constant");
continue;
}
if (count<maxsw-2) {
value[count]=val[0];
printlabel(label[count++]=getlabel());
}
else error("Too many case statments");
if (match(":")==0) error("Missing colon");
}
else if (amatch("default",7)) {
end_label=getlabel();
printlabel(end_label);
if (match(":")==0) error("Missing colon");
}
else if (match("}")) {
jump(que[wqlab]);	/* jump past switch data */
temp=0;
printlabel(label_switch);
eswitch(label_switch=getlabel());
printlabel(label_switch);
while(temp<count) {
defwrd(value[temp],0);
defwrd(label[temp++],1);
}
defwrd(end_label,1);defwrd(0,0);
delwhile();
printlabel(que[wqlab]);
locptr=que[wqsym];
Zsp=modstk(que[wqsp]);
return;
}
else statement();
}}
/* all done */
